package edu.washington.cs.radishsim;

import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

import static edu.washington.cs.radishsim.MemorySystemConstants.*;

/** Base class for all cache lines */
abstract class CoherentLine {
   protected LineAddress addr;

   abstract public boolean valid();

   abstract public boolean invalid();

   abstract public void invalidate();

   abstract public LineAddress lineAddress();

   abstract public boolean contains(ByteAddress ba);
}

/** A simple valid/invalid coherence scheme */
class VILine extends CoherentLine {
   protected boolean valid = false;

   public VILine() {}

   public VILine(LineAddress a) {
      this.valid = true;
      this.addr = a;
   }

   public boolean valid() {
      return valid;
   }

   public boolean invalid() {
      return !valid();
   }

   public void invalidate() {
      valid = false;
   }

   public LineAddress lineAddress() {
      return addr;
   }

   public boolean contains(ByteAddress ba) {
      /*
       * NB: use this.addr.lineAddress().equals() to dynamic dispatch to the right kind of
       * comparison. E.g., with a MetadataByteAddress, we will do an ownership-aware metadata
       * comparison, and with a DataByteAddress we will match with data addresses only.
       */
      return this.addr.equals(ba.lineAddress());
   }

   @Override
   public String toString() {
      return "addr=" + (addr == null ? "null" : addr.toString()) + (valid ? " valid" : " invalid");
   }
}

enum MESIState {
   MESI_MODIFIED, MESI_EXCLUSIVE, MESI_SHARED, MESI_INVALID
};

/** MESI coherence */
class MESILine extends VILine {
   protected MESIState state = MESIState.MESI_INVALID;

   public MESILine() {
      super();
   }

   public MESILine(LineAddress a) {
      // NB: super(a) sets valid = true;
      super(a);
   }

   public MESIState getState() {
      return state;
   }

   public void changeStateTo(MESIState s) {
      valid = (s != MESIState.MESI_INVALID);
      state = s;
   }

   @Override
   public boolean valid() {
      return this.valid && (this.state != MESIState.MESI_INVALID);
   }

   @Override
   public void invalidate() {
      this.valid = false;
      this.state = MESIState.MESI_INVALID;
   }

   @Override
   public String toString() {
      return super.toString() + " " + this.state.toString();
   }
};

/** Radish cache lines */
class RadishLine extends MESILine {

   enum Compacted {
      EXPANDED, COMPACT
   };

   /** Whether this line consists of metadata or regular data */
   protected boolean                  metadata = false;

   /** Whether this line consists of compact metadata or not. */
   protected Compacted                compact;

   /** the permissions this core has */
   protected MemoryAccessPermission[] localPerms;

   protected EpochOffset[]            localReadEpochOffsets;
   protected EpochOffset[]            localWriteEpochOffsets;

   /** Whether this byte is a leaf read or not. */
   protected boolean[]                leafRead;

   /**
    * whether this is a byte access or not. Used by the type safety optimization so that we only GC
    * clocks for byte data.
    */
   protected boolean[]                byteSizedAccess;

   /**
    * A bit is set to mark the first byte of a (potentially multi-byte) program variable. The number
    * of set bits is the number of program variables in this line.
    */
   protected boolean[]                programVariableStarts;

   private void init(Compacted c) {
      final int SIZE;

      switch ( c ) {
      case COMPACT:
         SIZE = MemorySystemConstants.LINE_SIZE()
                / MemorySystemConstants.COMPACT_METADATA_BYTES_PER_DATA_BYTE();
         break;
      case EXPANDED:
         SIZE = MemorySystemConstants.LINE_SIZE()
                / MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE();
         break;
      default:
         throw new IllegalArgumentException();
      }

      this.compact = c;
      localPerms = new MemoryAccessPermission[SIZE];
      localReadEpochOffsets = new EpochOffset[SIZE];
      localWriteEpochOffsets = new EpochOffset[SIZE];
      leafRead = new boolean[SIZE];
      byteSizedAccess = new boolean[SIZE];
      programVariableStarts = new boolean[SIZE];

      clearMetadata();
   }

   public RadishLine(Compacted c) {
      super();
      init(c);
   }

   public RadishLine(Compacted c, LineAddress a) {
      super(a);
      init(c);
      if ( c == Compacted.COMPACT ) {
         // ensure compact md lines are aligned properly
         assert MemorySystemConstants.COMPACT_MD_ALIGNED(a.lineAddr);
      }
   }

   @Override
   public boolean contains(ByteAddress ba) {
      // check ownership
      if ( metadata && ba instanceof MetadataByteAddress ) {
         MetadataByteAddress mba = (MetadataByteAddress) ba;
         MetadataLineAddress myLineAddr = (MetadataLineAddress) this.lineAddress();
         if ( !mba.owner().equals(myLineAddr.owner()) ) return false;
      }

      return containsUnowned(ba);
   }

   public boolean containsUnowned(ByteAddress ba) {
      if ( (ba instanceof DataByteAddress && metadata)
           || (ba instanceof MetadataByteAddress && !metadata) ) { //
         return false;
      }
      // types match up

      // range check
      final long base = this.lineAddress().get();
      long offset;
      if ( !metadata || isCompact() ) {
         offset = DATA_BYTES_PER_METADATA_LINE(Compacted.COMPACT);
      } else {
         offset = DATA_BYTES_PER_METADATA_LINE(Compacted.EXPANDED);
      }

      if ( metadata ) {
         offset *= EXPANDED_METADATA_BYTES_PER_DATA_BYTE();
      }

      return ba.get() >= base && ba.get() < (base + offset);
   }

   @Override
   public String toString() {
      return super.toString() + (metadata ? " md" : "") + (isCompact() ? " cmp" : " exp");
   }

   /** @Override */
   public void changeStateTo(MESIState newstate) {
      // metadata should always be in the Modified state
      if ( metadata ) {
         assert (MESIState.MESI_MODIFIED == newstate);
      }
      super.changeStateTo(newstate);
   }

   public void setMetadata() {
      clearMetadata();
      metadata = true;
      assert (MESIState.MESI_MODIFIED == state);
   }

   public void clearMetadata() {
      metadata = false;
      for ( int i = 0; i < localPerms.length; i++ ) {
         localPerms[i] = MemoryAccessPermission.NOTHING_OK;
      }
      resetLocalEpochOffsets();
   }

   public boolean isMetadata() {
      return metadata;
   }

   public boolean isCompact() {
      return compact == Compacted.COMPACT;
   }

   public boolean isExpanded() {
      return compact == Compacted.EXPANDED;
   }

   /** Record the local access as occurring during the given epoch. */
   public List<EpochOffset> recordLocalAccess(final MetadataAccess access, final EpochOffset eo,
                                              final boolean byteAccessesOnly,
                                              final boolean writesClobberReads) {
      assert metadata;
      assert eo.touched();

      final short i = access.metadataLineOffset(compact);
      programVariableStarts[i] = true;
      for ( int j = i+1; j < i+access.size(); j++ ) {
         programVariableStarts[j] = false;
      }

      List<EpochOffset> freedEpochOffsets = new LinkedList<EpochOffset>();

      for ( MetadataAccess a = new MetadataAccess(access); a.keepGoing(); a.incr() ) {
         recordLocalAccess(access.type(), a.addr(), eo, access.size() == 1, freedEpochOffsets,
                           byteAccessesOnly, writesClobberReads);
      }

      return freedEpochOffsets;
   }

   /**
    * Record the local access as occurring during the given epoch.
    * 
    * @param writesClobberReads
    *           whether updates to write md should clobber read md
    */
   public void recordLocalAccess(final MemoryAccessType mtype, final MetadataByteAddress addr,
                                 final EpochOffset eo, final boolean byteSized,
                                 final List<EpochOffset> toFree, final boolean byteAccessesOnly,
                                 final boolean writesClobberReads) {
      assert metadata;
      assert eo.touched();
      if ( RadishSim.XASSERTS ) assert containsUnowned(addr);
      assert !(byteSized && isCompact()); // md should have been expanded beforehand

      short i = addr.metadataLineOffset(compact);

      // determine what epoch offsets need to be freed

      if ( byteAccessesOnly && byteSizedAccess[i] && !byteSized ) {
         // this was a byte location but isn't anymore, so free all byte-oriented md
         if ( localWriteEpochOffsets[i].touched() ) toFree.add(localWriteEpochOffsets[i]);
         if ( localReadEpochOffsets[i].touched() ) toFree.add(localReadEpochOffsets[i]);
         // TODO: this is problematic because we aren't updating IHS...
         localReadEpochOffsets[i] = localWriteEpochOffsets[i] = EpochOffset.UNTOUCHED_EPOCH_OFFSET;

      } else if ( byteAccessesOnly && !byteSizedAccess[i] && byteSized ) {
         /*
          * This wasn't a byte location before but is now, so clear previous non-byte-oriented md.
          * NB: don't *free* this md, since it was never allocated (since we were not a byte
          * location before).
          */
         localReadEpochOffsets[i] = localWriteEpochOffsets[i] = EpochOffset.UNTOUCHED_EPOCH_OFFSET;

      } else if ( !byteAccessesOnly || byteSizedAccess[i] ) {
         switch ( mtype ) {
         case WRITE:
            if ( localWriteEpochOffsets[i].touched() ) toFree.add(localWriteEpochOffsets[i]);
            if ( !writesClobberReads ) break;
            // NB: no break here, since writes clobber read and write md
         case READ:
            if ( localReadEpochOffsets[i].touched() ) toFree.add(localReadEpochOffsets[i]);
            break;
         }
      }

      // clear any previous byte access...
      byteSizedAccess[i] = false;

      switch ( mtype ) {
      case WRITE:
         localWriteEpochOffsets[i] = eo;
         if ( writesClobberReads ) localReadEpochOffsets[i] = EpochOffset.UNTOUCHED_EPOCH_OFFSET;
         break;
      case READ:
         localReadEpochOffsets[i] = eo;
         // all reads are leaf reads, initially
         leafRead[i] = true;
         break;
      }

      // ...and record new byte access if appropriate
      if ( byteSized ) {
         byteSizedAccess[i] = true;
      }
   }

   /*
   public int numProgramVariables() {
      int vars = 0;
      for ( boolean b : programVariableStarts ) {
         if (b) vars++;
      }
      return vars; 
   }
   */
   
   public boolean startOfProgramVariable(final MetadataByteAddress addr) {
      short i = addr.metadataLineOffset(compact);
      return programVariableStarts[i];
   }
    

   public class ByteMetadata {
      EpochOffset epochOffset;
      boolean     byteSized;
      boolean     leafRead;
   }

   /**
    * Returns the local read/write epoch offset for the given byte.
    * 
    * @param mtype
    *           whether the memory access is a read or a write
    * @param a
    *           the address whose local epoch should be returned
    * @param byteAccessOnly
    *           return a valid epoch only if this byte has been accessed with an explicit byte
    *           access
    */
   public ByteMetadata getLocalEpochOffset(MemoryAccessType mtype, final MetadataByteAddress a) {
      assert metadata;
      if ( RadishSim.XASSERTS ) assert containsUnowned(a);

      ByteMetadata md = new ByteMetadata();
      short i = a.metadataLineOffset(compact);
      md.epochOffset = (mtype == MemoryAccessType.WRITE) ? localWriteEpochOffsets[i]
                                                        : localReadEpochOffsets[i];
      md.byteSized = byteSizedAccess[i];
      md.leafRead = (mtype == MemoryAccessType.READ && leafRead[i]);

      return md;
   }

   interface EpochOffsetVisitor {
      public void visit(MemoryAccessType mtype, MetadataByteAddress mba, ByteMetadata md);
   }

   public void visitLocalEpochOffsets(EpochOffsetVisitor v) {
      assert metadata;
      MetadataLineAddress mla = (MetadataLineAddress) this.addr;
      MetadataByteAddress mba = new MetadataByteAddress(mla);

      // NB: conservatively size the md access to handle compact lines
      // (which cover more data bytes)
      MetadataAccess ma = new MetadataAccess(
                                             mba,
                                             MemorySystemConstants.DATA_BYTES_PER_METADATA_LINE(RadishLine.Compacted.COMPACT));
      ByteMetadata md = new ByteMetadata();

      for ( int i = 0; i < this.localReadEpochOffsets.length; i++, ma.incr() ) {
         md.byteSized = this.byteSizedAccess[i];
         md.epochOffset = this.localReadEpochOffsets[i];
         md.leafRead = this.leafRead[i];
         v.visit(MemoryAccessType.READ, ma.addr(), md);

         md.epochOffset = this.localWriteEpochOffsets[i];
         md.leafRead = false;
         v.visit(MemoryAccessType.WRITE, ma.addr(), md);
      }
   }

   /**
    * Invalidate local r/w epochs for the given range of bytes
    * 
    * @param freedIDs
    *           a caller-supplied list to which free'd epoch offsets should be added
    * @param byteAccessesOnly
    *           only free offsets for explicit byte accesses
    * @return a list of the epoch offsets that should be free'd
    */
   public List<EpochOffset> invalidateLocalEpochs(final MetadataAccess access,
                                                  final boolean byteAccessesOnly) {
      assert metadata;
      if ( RadishSim.XASSERTS ) assert containsUnowned(access.addr());

      List<EpochOffset> freedEpochOffsets = new LinkedList<EpochOffset>();

      for ( MetadataAccess a = new MetadataAccess(access); a.keepGoing(); a.incr() ) {
         short i = a.addr().metadataLineOffset(compact);

         if ( !byteAccessesOnly || this.byteSizedAccess[i] ) {
            if ( this.localReadEpochOffsets[i].touched() )
               freedEpochOffsets.add(this.localReadEpochOffsets[i]);
            if ( this.localWriteEpochOffsets[i].touched() )
               freedEpochOffsets.add(this.localWriteEpochOffsets[i]);
         }

         this.localReadEpochOffsets[i] = this.localWriteEpochOffsets[i] = EpochOffset.UNTOUCHED_EPOCH_OFFSET;
         this.byteSizedAccess[i] = this.leafRead[i] = false;
      }

      return freedEpochOffsets;
   }

   public void invalidateLeafReads(final MetadataByteAddress mba) {
      assert metadata;
      if ( RadishSim.XASSERTS ) assert containsUnowned(mba);

      short i = mba.metadataLineOffset(compact);
      this.leafRead[i] = false;
   }

   /**
    * Returns true iff the eviction of this line would cause an in-hardware status downgrade, i.e.
    * this line records writes and/or leaf reads.
    */
   public boolean evictionWouldDowngradeInHardwareStatus() {
      assert this.metadata;

      for ( int i = 0; i < this.localReadEpochOffsets.length; i++ ) {
         if ( this.localWriteEpochOffsets[i].touched()
              || (this.localReadEpochOffsets[i].touched() && this.leafRead[i]) ) { return true; }
      }

      return false;
   }

   /** Reset all local epochs in this line back to "untouched" */
   public void resetLocalEpochOffsets() {
      // don't assert( metadata ) since this function is called from clearMetadata()
      for ( int i = 0; i < localReadEpochOffsets.length; i++ ) {
         localReadEpochOffsets[i] = localWriteEpochOffsets[i] = EpochOffset.UNTOUCHED_EPOCH_OFFSET;
         byteSizedAccess[i] = leafRead[i] = programVariableStarts[i] = false;
      }
   }

   /**
    * @param lperms
    *           The permissions to assign to this line. There may be perms for an expanded line or a
    *           compact line's worth of md.
    * @param startMla
    *           The mla of the starting byte of the perms.
    */
   public void assignLocalPerms(final MemoryAccessPermission[] lperms,
                                final MetadataLineAddress startMla) {
      // should never assign compact perms to an expanded line
      // assert lperms.length <= this.localPerms.length;
      if ( RadishSim.XASSERTS ) assert containsUnowned(new MetadataByteAddress(startMla));

      for ( int i = 0; i < lperms.length; i++ ) {
         assert lperms[i] != null;
      }

      if ( lperms.length == localPerms.length ) {
         localPerms = lperms;

      } else {
         final int start, end;

         if ( lperms.length > localPerms.length ) {
            assert !this.isCompact();
            assert lperms.length == (localPerms.length * 2);
            // do we take the first or second half of lperms?
            if ( startMla.get() == addr.get() ) {
               start = 0;
               end = DATA_BYTES_PER_METADATA_LINE(Compacted.EXPANDED);
            } else {
               start = DATA_BYTES_PER_METADATA_LINE(Compacted.EXPANDED);
               end = lperms.length;
            }
            int i = start;
            for ( int j = 0; i < end; i++, j++ ) {
               localPerms[j] = lperms[i];
            }

         } else { // we're compact but we only have perms for half
            assert this.isCompact();
            assert lperms.length == (this.localPerms.length / 2);
            // do we set the perms for the 1st half or the 2nd half?
            if ( startMla.get() == addr.get() ) {
               start = 0;
               end = DATA_BYTES_PER_METADATA_LINE(Compacted.EXPANDED);
            } else {
               start = DATA_BYTES_PER_METADATA_LINE(Compacted.EXPANDED);
               end = this.localPerms.length;
            }
            int i = start;
            for ( int j = 0; i < end; i++, j++ ) {
               localPerms[i] = lperms[j];
            }
         }

      }

      for ( int i = 0; i < this.localPerms.length; i++ ) {
         assert this.localPerms[i] != null;
      }

   }

   /** allow the given perms on the given range */
   public void setLocalPerms(MemoryAccessPermission perm, final MetadataAccess access) {
      assert metadata;
      assert perm != null;
      if ( RadishSim.XASSERTS ) assert containsUnowned(access.addr());

      for ( MetadataAccess a = new MetadataAccess(access); a.keepGoing(); a.incr() ) {
         localPerms[a.addr().metadataLineOffset(compact)] = perm;
      }
   }

   /** downgrade the perms on the given range so that they are no higher than perm */
   public void downgradeLocalPerms(MemoryAccessPermission perm, final MetadataAccess access) {
      assert metadata;
      assert perm != null;
      if ( RadishSim.XASSERTS ) assert containsUnowned(access.addr());

      for ( MetadataAccess a = new MetadataAccess(access); a.keepGoing(); a.incr() ) {
         int mlo = a.addr().metadataLineOffset(compact);
         localPerms[mlo] = MemoryAccessPermission.Compare.max(perm, localPerms[mlo]);
      }
   }

   /**
    * Checks for a data race on the given bytes. A race occurs if a local access L happens after
    * lastSyncEpoch (indicating there has been no sync to guard L).
    * 
    * @param access
    *           the bytes of the access being checked
    * @param lastSyncEpoch
    *           lastSyncEpoch is the last epoch that was well-synchronized between the caller and
    *           this cache.
    * @param base
    *           the base for the epoch offsets in this line
    * @return true if a race occurred, false otherwise
    */
   public boolean raceOccurred(final MetadataAccess access, Epoch lastSyncEpoch, EpochBase base) {
      assert metadata;
      if ( RadishSim.XASSERTS ) assert containsUnowned(access.addr());

      for ( MetadataAccess a = new MetadataAccess(access); a.keepGoing(); a.incr() ) {
         short i = a.addr().metadataLineOffset(compact);
         // only writes check local read epochs
         if ( a.write() && Epoch.Compare.gt(base.add(localReadEpochOffsets[i]), lastSyncEpoch) ) { return true; }
         // both reads and writes check local write epochs
         if ( Epoch.Compare.gt(base.add(localWriteEpochOffsets[i]), lastSyncEpoch) ) { return true; }
      }
      return false;
   }

   /**
    * @return true if reading the bytes in the given range is "ok" (no race), and false otherwise.
    */
   public boolean accessWouldBeRaceFree(final MetadataAccess access) {
      assert metadata;
      if ( RadishSim.XASSERTS ) assert containsUnowned(access.addr());

      boolean ok = true;
      final MemoryAccessPermission desired = access.read() ? MemoryAccessPermission.READ_OK
                                                          : MemoryAccessPermission.WRITE_OK;
      for ( MetadataAccess a = new MetadataAccess(access); a.keepGoing(); a.incr() ) {
         MemoryAccessPermission lp = localPerms[a.metadataLineOffset(compact)];
         ok &= MemoryAccessPermission.Compare.gte(lp, desired);
      }
      return ok;
   }

   /** Expand the metadata from this line into line l. */
   public void expand(RadishLine l) {
      assert this.isMetadata() && l.isMetadata() && this.isCompact();

      final int lStart = MemorySystemConstants.LINE_SIZE()
                         / MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE();
      final int lEnd = lStart * 2;

      // clear compact flag
      this.compact = Compacted.EXPANDED;
      assert l.compact == Compacted.EXPANDED;

      // copy local perms
      l.localPerms = Arrays.copyOfRange(this.localPerms, lStart, lEnd);
      this.localPerms = Arrays.copyOf(this.localPerms, lStart);
      for ( int i = 0; i < l.localPerms.length; i++ ) {
         assert l.localPerms[i] != null;
      }
      for ( int i = 0; i < this.localPerms.length; i++ ) {
         assert this.localPerms[i] != null;
      }

      // copy epochs
      l.localReadEpochOffsets = Arrays.copyOfRange(this.localReadEpochOffsets, lStart, lEnd);
      this.localReadEpochOffsets = Arrays.copyOf(this.localReadEpochOffsets, lStart);

      l.localWriteEpochOffsets = Arrays.copyOfRange(this.localWriteEpochOffsets, lStart, lEnd);
      this.localWriteEpochOffsets = Arrays.copyOf(this.localWriteEpochOffsets, lStart);

      // copy leaf reads
      l.leafRead = Arrays.copyOfRange(this.leafRead, lStart, lEnd);
      this.leafRead = Arrays.copyOf(this.leafRead, lStart);

      // copy byte-sized access information; shouldn't be any byte-sized accesses yet
      for ( int i = 0; i < this.byteSizedAccess.length; i++ ) {
         assert !this.byteSizedAccess[i];
      }
      l.byteSizedAccess = new boolean[lStart];
      this.byteSizedAccess = new boolean[lStart];
   }

};

/**
 * Used by HierarchicalCache to construct new line objects, since generic ctors can't be called
 * directly (due to type erasure).
 */
interface LineFactory<Line extends CoherentLine> {
   public Line create();

   public Line create(LineAddress la, RadishLine.Compacted c);
}