package edu.washington.cs.radishsim;

import java.io.IOException;
import java.io.Writer;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/** A class that manages the set of processors in the system. */
public class Machine<Line extends RadishLine> {

   private final Processor<Line>[]                  processors;
   private final CheckerProcessor<Line>[]           checkers;
   private final Map<CpuId, LinkedList<DataAccess>> queuedAccesses;

   private final MachineParams<Line>                params;

   public enum LeafReadConfiguration {
      PRECISE_LEAF_READS, LAZY_LEAF_READS, NO_LEAF_READS
   }

   public enum SimulationMode {
      BASELINE, RADISH, FASTTRACK
   }

   /**
    * Arguments to the Machine ctor. We encode these values as abstract methods so that we can't
    * forget to initialize one of them. The values get initialized by creating an anonymous subclass
    * that is forced to override all these methods.
    */
   static abstract class MachineParams<Line extends RadishLine> {
      /** Whether to simulate the Radish processor extensions or not */
      abstract SimulationMode simulationMode();

      /** The number of processors to simulate */
      abstract int numProcessors();

      /** The cache geometry for the private L1 caches */
      abstract CacheConfiguration<Line> l1config();

      /** Whether to simulate private L2s or not */
      abstract boolean useL2();

      /** The cache geometry for the private L2 caches */
      abstract CacheConfiguration<Line> l2config();

      /** Whether to simulate a shared L3 or not */
      abstract boolean useL3();

      /** The cache geometry for the shared L3 cache */
      abstract CacheConfiguration<Line> l3config();

      abstract LineFactory<Line> lineFactory();

      /** The number of bits of storage available for local epoch offsets */
      abstract int bitsPerLocalEpochOffset();

      abstract boolean useWholeLineSoftwareChecks();

      abstract boolean useClockGC();

      abstract boolean useTypeSafetyOpt();

      abstract boolean remoteAccessesAffectLRU();

      /**
       * The fixed costs of doing a software check: taking a snapshot of location and threads'
       * vector clocks, sending snapshot, etc.
       */
      abstract int swCheckCost();

      abstract int hwCheckCost();

      abstract int epochRolloverCost();

      abstract boolean evictUselessMetadata();

      abstract LeafReadConfiguration leafReads();

      abstract boolean metadataLLCEvictionPolicy();

      abstract int asyncProcessingCores();

      abstract boolean simulateSoftwareChecks();

      abstract boolean compactMetadata();

      abstract int cacheUsageSnapshotFrequencyInCycles();
   }

   /** shared by all processors */
   SoftwareMetadata globalMetaDataTable;

   /** Set of lines that contain byte-sized data. Shared by all processors. */
   //final static Set<DataLineAddress> linesWithByteData = new HashSet<DataLineAddress>();
   //Counter                           g_ByteContainingLinesAccesses;
   //Counter                           g_NonByteContainingLinesAccesses;

   @SuppressWarnings( "unchecked" )
   public Machine(MachineParams<Line> args) {
      
      /* HACK: see Counter.currentCpu for details */
      // map these counters to the fake cpu for machine-wide stats
      //Counter.currentCpu = new CpuId(-1);
      //g_ByteContainingLinesAccesses = new Counter("g_ByteContainingLinesAccesses");
      //g_NonByteContainingLinesAccesses = new Counter("g_NonByteContainingLinesAccesses");
      //Counter.currentCpu = null;

      this.params = args;

      Map<UnownedMetadataByteAddress, InHardwareStatus> inHwStatus = new HashMap<UnownedMetadataByteAddress, InHardwareStatus>();
      Map<LineAddress, Integer> varmap = new HashMap<LineAddress, Integer>();
      globalMetaDataTable = new SoftwareMetadata(args.numProcessors());
      Map<Long, SyncObject> syncObjectTable = new HashMap<Long, SyncObject>();
      checkers = new CheckerProcessor[args.asyncProcessingCores()];

      // construct processors
      processors = new Processor[args.numProcessors()];
      for ( int i = 0; i < processors.length; i++ ) {
         CpuId cpuid = new CpuId(i);
         /* HACK: see Counter.currentCpu for details */
         Counter.currentCpu = cpuid;
         processors[i] = new Processor<Line>(args, this, cpuid, globalMetaDataTable,
                                             syncObjectTable, processors, inHwStatus, varmap,
                                             checkers);
      }
      Counter.currentCpu = null;

      // construct checker processors
      for ( int i = 0; i < checkers.length; i++ ) {
         CpuId checkerCpuid = new CpuId(processors.length + i);
         /* HACK: see Counter.currentCpu for details */
         Counter.currentCpu = checkerCpuid;
         checkers[i] = new CheckerProcessor<Line>(args, this, checkerCpuid, globalMetaDataTable,
                                                  syncObjectTable, processors, inHwStatus, varmap,
                                                  checkers);
      }
      Counter.currentCpu = null;

      queuedAccesses = new HashMap<CpuId, LinkedList<DataAccess>>();
   }

   /** map from thread id to cpu id */
   public CpuId cpuOfTid(ThreadId tid) {
      // NB: we have a very simple mapping of threads onto caches
      return processors[tid.get() % processors.length].id;
   }

   Processor<Line> getProc(CpuId cpuid) {
      Processor<Line> p = processors[cpuid.get()];
      assert p != null;
      return p;
   }

   CheckerProcessor<Line> getCheckerProc(CpuId cpuid) {
      assert cpuid.get() >= processors.length;
      CheckerProcessor<Line> cp = checkers[cpuid.get() - processors.length];
      assert cp != null;
      return cp;
   }

   public void dumpStats(Writer wr, String prefix, String suffix) throws IOException {
      for ( Processor<Line> p : processors ) {
         p.finalizeCounters();
      }
      // the Counter class keeps track of all its instances, so we only need to dump once
      Counter.dumpCounters(wr, prefix, suffix);
      globalMetaDataTable.dump(wr, prefix, suffix);
      processors[0].dumpCacheUsageSnapshots(wr, prefix, suffix);
   }

   /** Enqueue some non-metadata accesses, to be executed after the current access completes. */
   void enqueueAccesses(CpuId cpu, List<DataAccess> accesses) {
      if ( accesses == null ) return;

      LinkedList<DataAccess> das = queuedAccesses.get(cpu);
      if ( das == null ) {
         das = new LinkedList<DataAccess>(accesses);
         queuedAccesses.put(cpu, das);
      }
      das.addAll(accesses); // append
   }

   private void processQueuedAccesses() {
      for ( Map.Entry<CpuId, LinkedList<DataAccess>> mei : queuedAccesses.entrySet() ) {
         for ( DataAccess da : mei.getValue() ) {
            cacheAccess(mei.getKey(), da.write(), da.addr().get(), da.size(), false);
         }
      }
      queuedAccesses.clear();
   }

   public void insnsExecuted(final CpuId cpuid, byte n) {
      getProc(cpuid).insnsExecuted(n);
   }

   public void cacheRead(final CpuId cpuid, final long addr, final int size) {
      cacheAccess(cpuid, false, addr, size, true);
   }

   public void cacheRead(final CpuId cpuid, final long addr, final int size,
                         final boolean doMetadataAccess) {
      cacheAccess(cpuid, false, addr, size, doMetadataAccess);
   }

   public void cacheWrite(final CpuId cpuid, final long addr, final int size) {
      cacheAccess(cpuid, true, addr, size, true);
   }

   public void cacheWrite(final CpuId cpuid, final long addr, final int size,
                          final boolean doMetadataAccess) {
      cacheAccess(cpuid, true, addr, size, doMetadataAccess);
   }

   public void cacheAccess(final CpuId cpuid, final boolean write, final long addr, final int size,
                           boolean doMetadataAccess) {

      Processor<Line> proc = getProc(cpuid);
      // size things conservatively, assuming expanded metadata
      final int MD_LINE_SIZE = MemorySystemConstants.LINE_SIZE()
                               / MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE();

      Processor.DataMemoryAccessResult mopResult = new Processor.DataMemoryAccessResult();
      mopResult.remoteCommunicatedHappened = false;
      int remainingSize = size;

      for ( long a = addr; remainingSize > 0; ) {
         DataByteAddress dba = new DataByteAddress(a);
         int data_bytesFromStartOfLine = dba.lineOffset();
         int data_maxSizeAccessWithinThisLine = MemorySystemConstants.LINE_SIZE()
                                                - data_bytesFromStartOfLine;

         // data access
         int accessSize = Math.min(remainingSize, data_maxSizeAccessWithinThisLine);
         Processor.DataMemoryAccessResult tempMor;
         if ( write ) {
            tempMor = proc.write(new DataAccess(MemoryAccessType.WRITE, dba, accessSize));
         } else {
            tempMor = proc.read(new DataAccess(MemoryAccessType.READ, dba, accessSize));
         }
         mopResult.aggregate(tempMor);

         a += accessSize;
         remainingSize -= accessSize;
      }

      switch ( params.simulationMode() ) {
      case BASELINE:
         return;
      case FASTTRACK:
         if ( !doMetadataAccess ) return;
         MetadataByteAddress mba = new MetadataByteAddress(addr, cpuid);
         // NB: we only do a FT check for the first byte of the access; we assume all
         // accesses are the same.
         Processor<Line>.SwRaceCheckResult res;
         MetadataAccess ma = new MetadataAccess(write ? MemoryAccessType.WRITE
                                                     : MemoryAccessType.READ, mba, 1);
         res = proc.fastTrackRaceCheck(ma);
         enqueueAccesses(cpuid, res.swCheckAccesses);
         processQueuedAccesses();
         proc.stats.pc_ExecDrivenCycleCount.incr(MemorySystemConstants.FASTTRACK_CHECK_INSN_COST);
         return;
      case RADISH:
         if ( !doMetadataAccess ) return;
         break; // perform Radish metadata access (below)
      default:
         assert false;
      }

      /*
       * Metadata accesses are funky because the line size is smaller due to the
       * EXPANDED_METADATA_BYTES_PER_DATA_BYTE ratio. This smaller line size is enforced by the
       * lines as well. So, we have to compensate by generating scaled-down access sizes, even
       * though the number of addresses we touch is scaled up.
       */
      int remainingDataBytes = size;
      for ( long a = addr; remainingDataBytes > 0; ) {
         // NB: a is, effectively, a DataByteAddress
         MetadataByteAddress mba = new MetadataByteAddress(a, proc.id);
         // db_ means units are data bytes (not metadata bytes)
         int db_bytesFromStartOfLine = mba.metadataLineOffset(RadishLine.Compacted.EXPANDED);
         int db_maxSizeAccessWithinThisLine = MD_LINE_SIZE - db_bytesFromStartOfLine;

         final int db_accessSize = Math.min(remainingDataBytes, db_maxSizeAccessWithinThisLine);
         MetadataAccess mdAccess = new MetadataAccess(write ? MemoryAccessType.WRITE
                                                           : MemoryAccessType.READ, mba,
                                                      db_accessSize);
         proc.accessMetadata(mdAccess, mopResult);

         a += db_accessSize;
         remainingDataBytes -= db_accessSize;
      }

      // recurse!
      processQueuedAccesses();

   }

   /**
    * @param performingCpu
    *           the thread performing the sync op
    * @param op
    *           whether this is a source or a sink in the HB graph
    * @param validSource
    *           whether there is a valid source for this sink op (e.g. the first lock acquire has no
    *           "source")
    * @param syncObject
    *           the object being synchronized on
    */
   public void syncOp(final CpuId performingCpu, SynchronizationOperationType op,
                      final boolean validSource, final long syncObject) {
      getProc(performingCpu).syncOp(op, validSource, syncObject);
   }
};
