package edu.washington.cs.radishsim;

import static org.junit.Assert.*;

import java.util.Random;

import org.junit.Before;
import org.junit.Test;

public class CompactMetadataTests {
   static Machine<RadishLine>            m;

   static final CpuId                    P0                   = new CpuId(0);
   static final CpuId                    P1                   = new CpuId(1);
   static final int                      CORES                = 2;
   static final int                      LINE_SIZE            = 16;
   static final int                      L1_ASSOC             = 2;
   static final int                      L1_CACHE_SIZE        = 64;          // 4 lines, 2 sets
   // (fully associative)
   static final int                      BITS_PER_LOCAL_EPOCH = 2;

   static {
      MemorySystemConstants.unsafeSetLineSize(LINE_SIZE);
      assert RadishSim.XASSERTS;
   }

   static CacheConfiguration<RadishLine> l1config             = new CacheConfiguration<RadishLine>() {
                                                                 {
                                                                    lineSize = LINE_SIZE;
                                                                    cacheSize = L1_CACHE_SIZE;
                                                                    assoc = L1_ASSOC;
                                                                    level = CacheLevel.L1;
                                                                 }
                                                              };

   @Before
   public void setUp() {
      Machine.MachineParams<RadishLine> params = new DefaultTestingMachineParams() {

         @Override
         int numProcessors() {
            return CORES;
         }

         @Override
         int bitsPerLocalEpochOffset() {
            return BITS_PER_LOCAL_EPOCH;
         }

         @Override
         CacheConfiguration<RadishLine> l1config() {
            return l1config;
         }

         @Override
         LineFactory<RadishLine> lineFactory() {
            return new LineFactory<RadishLine>() {
               @Override
               public RadishLine create() {
                  return new RadishLine(RadishLine.Compacted.COMPACT);
               }

               @Override
               public RadishLine create(LineAddress la, RadishLine.Compacted c) {
                  return RadishSim.createRadishLine(la, c);
               }
            };
         }

         @Override
         boolean compactMetadata() {
            return true;
         }

      };
      m = new Machine<RadishLine>(params);
   }

   boolean compact(CpuId cpu, long addr) {
      MetadataByteAddress mba = new MetadataByteAddress(addr, cpu);
      MemoryResponse<RadishLine> resp = m.getProc(cpu).L1cache.ownedMdSearch(mba, false);
      assertNotNull(resp.lineHit);
      assertTrue(resp.lineHit.isMetadata());
      return resp.lineHit.isCompact();
   }

   boolean expanded(CpuId cpu, long addr) {
      MetadataByteAddress mba = new MetadataByteAddress(addr, cpu);
      MemoryResponse<RadishLine> resp = m.getProc(cpu).L1cache.ownedMdSearch(mba, false);
      assertNotNull(resp.lineHit);
      return resp.lineHit.isExpanded();
   }

   // jargon:
   // Ad means data at location A
   // Amd=B means the metadata for data location A resides at location B

   @Test
   public void test_basicCompact() {
      Processor<RadishLine> p0 = m.getProc(P0);

      // fetch 0md=0 md line in compact mode
      m.cacheRead(P0, 0, 2);
      assertEquals(1, p0.stats.pc_l1md.pc_ReadMisses.get());
      assertTrue(compact(P0, 0));

      // touch last 2 bytes in the data line
      m.cacheRead(P0, LINE_SIZE - 2, 2);
      assertEquals(1, p0.stats.pc_l1md.pc_ReadMisses.get());
      assertTrue(compact(P0, (LINE_SIZE - 2) & MemorySystemConstants.COMPACT_MD_ALIGNMENT_MASK()));
   }

   @Test
   public void test_expandOnFetch() {
      Processor<RadishLine> p0 = m.getProc(P0);

      // fetch 0md=0 md line in expanded mode
      m.cacheRead(P0, 0, 1);
      assertEquals(1, p0.stats.pc_CompactMdCompanionChecks.get());
      assertTrue(expanded(P0, 0));
   }

   @Test
   public void test_expandAfterFetch() {
      Processor<RadishLine> p0 = m.getProc(P0);

      // fetch 0md=0 md line in compact mode
      m.cacheRead(P0, 0, 2);
      assertEquals(1, p0.stats.pc_l1md.pc_ReadMisses.get());
      assertTrue(compact(P0, 0));

      // touch last byte in the data line, triggers expansion
      m.cacheRead(P0, LINE_SIZE - 1, 1);
      assertEquals(2, p0.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(1, p0.stats.pc_CompactMdExpansions.get());
      assertTrue(expanded(P0, 0));
      assertTrue(expanded(P0, LINE_SIZE - 1));
   }

   @Test
   public void test_bottomLineExpanded() {
      Processor<RadishLine> p0 = m.getProc(P0);

      // fetch 0md=0 md line (bottom) in expanded mode
      m.cacheRead(P0, 0, 1);
      assertEquals(1, p0.stats.pc_l1md.pc_ReadMisses.get());
      assertTrue(expanded(P0, 0));

      // touch last byte in the data line, fetch top
      m.cacheRead(P0, LINE_SIZE - 2, 2);
      assertEquals(2, p0.stats.pc_l1md.pc_ReadMisses.get());
      assertTrue(expanded(P0, LINE_SIZE-2));
   }

   @Test
   public void test_topLineExpanded() {
      Processor<RadishLine> p0 = m.getProc(P0);

      // touch last byte in the data line, fetch top expanded
      m.cacheRead(P0, LINE_SIZE - 2, 1);
      assertEquals(1, p0.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(1, p0.stats.pc_CompactMdCompanionChecks.get());
      assertTrue(expanded(P0, LINE_SIZE-2));

      // fetch 0md=0 md line (bottom) in expanded mode
      m.cacheRead(P0, 0, 2);
      assertEquals(2, p0.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(2, p0.stats.pc_CompactMdCompanionChecks.get());
      assertTrue(expanded(P0, 0));
   }

   @Test
   public void test_randomAccesses() {
      Random rng = new Random(123);

      for ( int i = 0; i < 50000; i++ ) {
         int addr = rng.nextInt() % 256;
         boolean proc = rng.nextBoolean();
         CpuId cpu = proc ? P0 : P1;
         boolean read = rng.nextBoolean();

         boolean doSyncOp = (rng.nextInt() % 10) == 0;
         if ( doSyncOp )
            m.syncOp(cpu, read ? SynchronizationOperationType.SOURCE
                              : SynchronizationOperationType.SOURCE, false, addr);

         int size = rng.nextBoolean() ? 1 : 2;
         if ( read ) m.cacheRead(cpu, addr, size);
         else m.cacheWrite(cpu, addr, size);
      }
   }

}
