package edu.washington.cs.radishsim;

import static org.junit.Assert.*;

import org.junit.Before;
import org.junit.Test;

public class MultiproTests {

   static Machine<RadishLine>            m;

   static final CpuId                    P0                   = new CpuId(0);
   static final CpuId                    P1                   = new CpuId(1);
   static final int                      CORES                = 2;
   static final int                      L1_ASSOC             = 2;
   static final int                      LINE_SIZE            = 4;
   static final int                      L1_CACHE_SIZE        = 16;          // 4 blocks, 2 sets
   static final int                      BITS_PER_LOCAL_EPOCH = 2;

   static {
      MemorySystemConstants.unsafeSetLineSize(LINE_SIZE);
      assert RadishSim.XASSERTS;
   }

   static CacheConfiguration<RadishLine> l1config             = new CacheConfiguration<RadishLine>() {
                                                                 {
                                                                    cacheSize = L1_CACHE_SIZE;
                                                                    lineSize = LINE_SIZE;
                                                                    assoc = L1_ASSOC;
                                                                    level = CacheLevel.L1;
                                                                 }
                                                              };

   @Before
   public void setUp() {
      Machine.MachineParams<RadishLine> params = new DefaultTestingMachineParams() {

         @Override
         int numProcessors() {
            return CORES;
         }

         @Override
         CacheConfiguration<RadishLine> l1config() {
            return l1config;
         }

         @Override
         int bitsPerLocalEpochOffset() {
            return BITS_PER_LOCAL_EPOCH;
         }
      };

      m = new Machine<RadishLine>(params);
   }

   // jargon:
   // Ad means data at location A
   // Amd=B means the metadata for data location A resides at location B

   @Test
   public void test_multiproSimple() {
      Processor<RadishLine> p = m.getProc(P0);

      // bring 0d,0md=0 into same set
      m.cacheRead(P0, 0, 1);
      assertEquals(1, p.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(0, p.stats.pc_MetadataEvictions.get());

      // fetch 16d,16md=32, which map to set 0 and evict 0d,0md=0
      m.cacheRead(P0, L1_CACHE_SIZE, 1);
      assertEquals(2, p.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(1, p.stats.pc_MetadataEvictions.get());

      // fetch 0d,0md=0 again, evicting 16d,16md=32
      m.cacheRead(P0, 0, 1);
      assertEquals(3, p.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(2, p.stats.pc_MetadataEvictions.get());
   }

   @Test
   public void test_metadataDoesntNeedCoherence() {
      Processor<RadishLine> p0 = m.getProc(P0);
      Processor<RadishLine> p1 = m.getProc(P1);

      // bring 0d and 0md=0 into set 0 at P0
      m.cacheRead(P0, 0, 1);
      assertEquals(1, p0.stats.pc_l1d.pc_ReadMisses.get());
      assertEquals(1, p0.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(0, p0.stats.pc_MetadataEvictions.get());

      // P1 writes to 0d, invalidating P0's 0d. P1 gets its own 0md=0
      m.cacheWrite(P1, 0, 1);
      assertEquals(1, p1.stats.pc_l1d.pc_WriteMisses.get());
      assertEquals(1, p1.stats.pc_WriteRemoteHits.get());
      assertEquals(1, p1.stats.pc_l1md.pc_WriteMisses.get());
      assertEquals(0, p1.stats.pc_MetadataEvictions.get());

      // P0 misses on 0d, but P0's 0md=0 is still valid
      m.cacheWrite(P0, 0, 1);
      assertEquals(1, p0.stats.pc_l1d.pc_WriteMisses.get());
      assertEquals(1, p0.stats.pc_WriteRemoteHits.get());
      assertEquals(1, p0.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(0, p0.stats.pc_MetadataEvictions.get());

      // P1 reads 0d, downgrading P0's 0d, but P1's 0md=0 is still valid
      m.cacheRead(P1, 0, 1);
      assertEquals(1, p1.stats.pc_l1d.pc_ReadMisses.get());
      assertEquals(1, p1.stats.pc_ReadRemoteHits.get());
      assertEquals(0, p1.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(1, p1.stats.pc_l1md.pc_WriteMisses.get());
      assertEquals(0, p1.stats.pc_MetadataEvictions.get());

      // P0 upgrade misses on 0d, but P0's 0md=0 is still valid
      m.cacheWrite(P0, 0, 1);
      assertEquals(1, p0.stats.pc_UpgradeMisses.get());
      assertEquals(1, p0.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(0, p0.stats.pc_MetadataEvictions.get());
   }

   @Test
   public void test_rwConflictViaRemoteBits() {
      Processor<RadishLine> p0 = m.getProc(P0);
      Processor<RadishLine> p1 = m.getProc(P1);
      UnownedMetadataByteAddress umba0 = new UnownedMetadataByteAddress(0);

      // bring 0d and 0md=0 into set 0 at P0
      m.cacheRead(P0, 0, 1);
      assertEquals(1, p0.stats.pc_LocalReadViolatesLocalPerms.get());
      assertEquals(0, p0.stats.pc_DataRaces.get());

      assertEquals(0, p0.stats.pc_HwCacheChecks.get());
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_HwCacheChecks.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());
      assertEquals(InHardwareStatus.EVERYTHING, p0.inHardwareStatus.get(umba0));

      // P1 writes to 0d, which fetches P0's read as remote bits.
      // There's a conflict on these remote bits
      assertEquals(0, p1.stats.pc_DataRaces.get());
      m.cacheWrite(P1, 0, 1);
      assertEquals(1, p1.stats.pc_LocalWriteViolatesLocalPerms.get());
      assertEquals(1, p1.stats.pc_DataRaces.get());

      assertEquals(0, p0.stats.pc_HwCacheChecks.get());
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_HwCacheChecks.get()); // P1's check piggybacks on the fill
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get()); // sufficient hw md
      assertEquals(InHardwareStatus.EVERYTHING, p0.inHardwareStatus.get(umba0));

      // P0 reads again, triggering another conflict
      m.cacheRead(P0, 0, 1);
      assertEquals(2, p0.stats.pc_LocalReadViolatesLocalPerms.get());
      assertEquals(1, p0.stats.pc_DataRaces.get());

      assertEquals(1, p0.stats.pc_HwCacheChecks.get());
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_HwCacheChecks.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());
      assertEquals(InHardwareStatus.EVERYTHING, p0.inHardwareStatus.get(umba0));
   }

   @Test
   public void test_wwConflictViaRemoteBits() {
      Processor<RadishLine> p0 = m.getProc(P0);
      Processor<RadishLine> p1 = m.getProc(P1);

      // bring 0d and 0md=0 into set 0 at P0
      m.cacheWrite(P0, 0, 1);
      assertEquals(1, p0.stats.pc_LocalWriteViolatesLocalPerms.get());
      assertEquals(0, p0.stats.pc_DataRaces.get());

      assertEquals(0, p0.stats.pc_HwCacheChecks.get());
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_HwCacheChecks.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());

      // P1 writes to 0d, which fetches P0's write as remote bits.
      // There's a conflict on these remote bits
      assertEquals(0, p1.stats.pc_DataRaces.get());
      m.cacheWrite(P1, 0, 1);
      assertEquals(1, p1.stats.pc_LocalWriteViolatesLocalPerms.get());
      assertEquals(1, p1.stats.pc_DataRaces.get());

      assertEquals(0, p0.stats.pc_HwCacheChecks.get());
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_HwCacheChecks.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());

      // P0 writes again, triggering another conflict
      m.cacheWrite(P0, 0, 1);
      assertEquals(2, p0.stats.pc_LocalWriteViolatesLocalPerms.get());
      assertEquals(1, p0.stats.pc_DataRaces.get());

      assertEquals(1, p0.stats.pc_HwCacheChecks.get());
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_HwCacheChecks.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());
   }

   @Test
   public void test_evictInvalidMetadata() {
      Processor<RadishLine> p0 = m.getProc(P0);
      Processor<RadishLine> p1 = m.getProc(P1);

      m.cacheRead(P0, 0, 1);
      assertEquals(0, p0.stats.pc_HwCacheChecks.get());
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_HwCacheChecks.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());

      // invalidates P0's 0md=0
      m.cacheWrite(P1, 0, 1);
      assertEquals(0, p0.stats.pc_HwCacheChecks.get());
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_HwCacheChecks.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());

      // evict 0md=0 from P0, but it doesn't have any valid md...
      m.cacheRead(P0, L1_CACHE_SIZE, 1);
      assertEquals(0, p0.stats.pc_MetadataEvictions.get());
      assertEquals(2, p0.stats.pc_SwVcChecksTotal.get());

      // so P0 doesn't need to do a sw check
      m.cacheWrite(P0, 0, 1);
      assertEquals(0, p0.stats.pc_HwCacheChecks.get());
      assertEquals(2, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_HwCacheChecks.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());
   }

   @Test
   public void test_writesAreValidMetadata() {
      Processor<RadishLine> p0 = m.getProc(P0);
      Processor<RadishLine> p1 = m.getProc(P1);

      m.cacheWrite(P0, 0, 1);
      assertEquals(0, p0.stats.pc_HwCacheChecks.get());
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_HwCacheChecks.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());

      // sufficient metadata: found a writer
      m.cacheRead(P1, 0, 1);
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());
      // assertEquals(1, p1.stats.pc_SufficientMd_HardwareMd.get());

      // no remote bits are set
      m.cacheRead(P1, 0, 1);
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());

      // P1 -> P0
      m.syncOp(P1, SynchronizationOperationType.SOURCE, false, 0xA);
      m.syncOp(P0, SynchronizationOperationType.SINK, true, 0xA);

      // no remote bits are set
      m.cacheRead(P0, 0, 1);
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());
   }

   @Test
   public void test_writeHitsInvalidateMetadata() {
      Processor<RadishLine> p0 = m.getProc(P0);
      Processor<RadishLine> p1 = m.getProc(P1);
      UnownedMetadataByteAddress umba0 = new UnownedMetadataByteAddress(0);

      m.cacheWrite(P0, 0, 1);
      assertEquals(0, p0.stats.pc_HwCacheChecks.get());
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_HwCacheChecks.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());
      assertEquals(InHardwareStatus.EVERYTHING, p0.inHardwareStatus.get(umba0));

      // P0 -> P1
      m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xA);
      m.syncOp(P1, SynchronizationOperationType.SINK, true, 0xA);

      // our write should invalidate P0's epochs/lperms
      m.cacheWrite(P1, 0, 1);
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_HwCacheChecks.get()); // piggy-back on miss
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());
      assertEquals(InHardwareStatus.EVERYTHING, p0.inHardwareStatus.get(umba0));
   }

   @Test
   public void test_preciseSoftwareVcChecks_NoPreviousAccesses() {
      Processor<RadishLine> p0 = m.getProc(P0);
      Processor<RadishLine> p1 = m.getProc(P1);

      // since there are no previous accesses, hw owns the md now
      m.cacheRead(P0, 0, 1);
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());

      // sufficient metadata: hw md
      m.cacheWrite(P1, 0, 1);
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());
      // assertEquals(1, p1.stats.pc_SufficientMd_HardwareMd.get());
   }

   @Test
   public void test_preciseSoftwareVcChecks_GlobalAccessPerms() {
      Processor<RadishLine> p0 = m.getProc(P0);
      Processor<RadishLine> p1 = m.getProc(P1);
      UnownedMetadataByteAddress umba0 = new UnownedMetadataByteAddress(0);

      // since there are no previous accesses, hw owns the md now
      m.cacheWrite(P0, 0, 1);
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());

      // evict 0md=0 from P0
      for ( int i = 1; i <= L1_ASSOC; i++ ) {
         m.cacheRead(P0, L1_CACHE_SIZE * i, 1);
      }
      assertEquals(1 + L1_ASSOC, p0.stats.pc_SwVcChecksTotal.get());
      assertNull(p0.inHardwareStatus.get(umba0));

      // P0 -> P1
      m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xA);
      m.syncOp(P1, SynchronizationOperationType.SINK, true, 0xA);

      // hw owns md again, and it's safe for everyone to read/write
      m.cacheRead(P0, 0, 1);
      assertEquals(2 + L1_ASSOC, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(InHardwareStatus.EVERYTHING, p0.inHardwareStatus.get(umba0));

      m.cacheRead(P1, 0, 1);
      assertEquals(2 + L1_ASSOC, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());
      assertEquals(InHardwareStatus.EVERYTHING, p0.inHardwareStatus.get(umba0));

      m.cacheWrite(P1, 0, 1);
      assertEquals(2 + L1_ASSOC, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());
      assertEquals(1, p1.stats.pc_DataRaces.get());
      assertEquals(InHardwareStatus.EVERYTHING, p0.inHardwareStatus.get(umba0));
   }

   @Test
   public void test_preciseSoftwareVcChecks_RestoreEpochs() {
      Processor<RadishLine> p0 = m.getProc(P0);
      Processor<RadishLine> p1 = m.getProc(P1);

      m.cacheRead(P0, 0, 1);
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());

      m.cacheWrite(P1, 0, 1);
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());

      // evict 0md=0 from all caches
      for ( int i = 1; i <= L1_ASSOC; i++ ) {
         m.cacheRead(P0, L1_CACHE_SIZE * i, 1);
         m.cacheRead(P1, L1_CACHE_SIZE * i, 1);
      }
      assertEquals(1 + L1_ASSOC, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get()); // piggy-backs off P0's checks

      // remember our read from before, no hw md since we don't h-a the previous writer (P1)
      // true, but then we set epochs again due to this read, which gives P1 h-a a reader
      // TODO: need a test that actually requires epochs to be set...are they actually redundant
      // with other mechanisms, though?
      m.cacheRead(P0, 0, 1);
      assertEquals(2 + L1_ASSOC, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());

      // P0 -> P1
      m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xA);
      m.syncOp(P1, SynchronizationOperationType.SINK, true, 0xA);

      // P1 h-a a previous reader
      m.cacheRead(P1, 0, 1);
      assertEquals(2 + L1_ASSOC, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());
   }

   @Test
   public void test_preciseSoftwareVcChecks_LastReaderWriter() {
      Processor<RadishLine> p0 = m.getProc(P0);
      Processor<RadishLine> p1 = m.getProc(P1);

      // last writer
      m.cacheWrite(P0, 0, 1);
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());

      // force epoch rollover
      for ( int i = 0; i < (1 << BITS_PER_LOCAL_EPOCH); i++ ) {
         m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xA);
      }
      assertEquals(1, p0.stats.pc_EpochRollovers.get());

      // evict 0md=0 from P0's cache
      for ( int i = 1; i <= L1_ASSOC; i++ ) {
         m.cacheRead(P0, L1_CACHE_SIZE * i, 1);
      }
      assertEquals(1 + L1_ASSOC, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());

      // last-reader-writer bit is set, because write epoch is unrepresentable
      m.cacheRead(P0, 0, 1);
      assertEquals(2 + L1_ASSOC, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());

      // P0 -> P1
      m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xB);
      m.syncOp(P1, SynchronizationOperationType.SINK, true, 0xB);

      // h-a last-reader-writer, no need for a sw check. sets in-hw bit
      m.cacheWrite(P1, 0, 1);
      assertEquals(2 + L1_ASSOC, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_DataRaces.get());

      // since in-hw bit is set, no need for a sw check here either
      m.cacheWrite(P0, 0, 1);
      assertEquals(2 + L1_ASSOC, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());
      assertEquals(1, p0.stats.pc_DataRaces.get());
   }

   @Test
   public void test_preciseSoftwareVcChecks_LastReaderWriterMultiByte() {
      Processor<RadishLine> p0 = m.getProc(P0);
      Processor<RadishLine> p1 = m.getProc(P1);

      // last writer
      m.cacheWrite(P0, 0, 2);
      assertEquals(1, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());

      // force epoch rollover
      for ( int i = 0; i < (1 << BITS_PER_LOCAL_EPOCH); i++ ) {
         m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xA);
      }
      assertEquals(1, p0.stats.pc_EpochRollovers.get());

      // evict 0md=0 from P0's cache
      for ( int i = 1; i <= L1_ASSOC; i++ ) {
         m.cacheRead(P0, L1_CACHE_SIZE * i, 1);
      }
      assertEquals(1 + L1_ASSOC, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());

      // last-reader-writer bit is set, because write epoch is unrepresentable
      m.cacheRead(P0, 0, 1);
      assertEquals(2 + L1_ASSOC, p0.stats.pc_SwVcChecksTotal.get());
      assertEquals(0, p1.stats.pc_SwVcChecksTotal.get());

      // P0 -> P1
      m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xB);
      m.syncOp(P1, SynchronizationOperationType.SINK, true, 0xB);

      // write adjacent byte. P1 h-a last-reader-writer, no need for a sw check. sets in-hw bit
      m.cacheWrite(P1, 1, 1);
      assertEquals(2 + L1_ASSOC, p0.stats.pc_SwVcChecksTotal.get());
      // assertEquals( 0, p1.stats.pc_SoftwareVCChecks.get() );
   }

}
