package edu.washington.cs.radishsim;

import static org.junit.Assert.assertEquals;

import org.junit.Before;
import org.junit.Test;

public class UniproTests {

   static Machine<RadishLine>            m;

   static final CpuId                    P0                   = new CpuId(0);
   static final int                      CORES                = 1;
   static final int                      L1_ASSOC             = 2;
   static final int                      LINE_SIZE            = 4;
   static final int                      L1_CACHE_SIZE        = 16;          // 4 blocks, 2 sets
   static final int                      BITS_PER_LOCAL_EPOCH = 1;

   static {
      MemorySystemConstants.unsafeSetLineSize(LINE_SIZE);
      assert RadishSim.XASSERTS;
   }

   static CacheConfiguration<RadishLine> l1config             = new CacheConfiguration<RadishLine>() {
                                                                 {
                                                                    cacheSize = L1_CACHE_SIZE;
                                                                    lineSize = LINE_SIZE;
                                                                    assoc = L1_ASSOC;
                                                                    level = CacheLevel.L1;
                                                                 }
                                                              };

   @Before
   public void setUp() {
      Machine.MachineParams<RadishLine> params = new DefaultTestingMachineParams() { 
         @Override
         int numProcessors() {
            return CORES;
         }
         
         @Override
         CacheConfiguration<RadishLine> l1config() {
            return l1config;
         }
         
         @Override
         int bitsPerLocalEpochOffset() {
            return BITS_PER_LOCAL_EPOCH;
         }
      };
      m = new Machine<RadishLine>(params);
   }

   // jargon:
   // Ad means data at location A
   // Amd=B means the metadata for data location A resides at location B

   @Test
   public void test_Counter() {
      Counter.currentCpu = P0;
      Counter c1 = new Counter("c1");
      Counter c2 = new Counter("c2");
      c1.incr();
      assertEquals(1, c1.get());
      c2.incr();
      assertEquals(1, c2.get());

      c1.set(12);
      assertEquals(12, c1.get());
      c1.set(12 + c2.get());
      assertEquals(13, c1.get());
   }

   @Test
   public void test_reads() {
      Processor<RadishLine> p = m.getProc(P0);

      // bring 0d,0md=0 into same set
      m.cacheRead(P0, 0, 1);
      assertEquals(1, p.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(0, p.stats.pc_MetadataEvictions.get());

      // fetch 16d,16md=32, which map to set 0 and evict 0d,0md=0
      m.cacheRead(P0, L1_CACHE_SIZE, 1);
      assertEquals(2, p.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(1, p.stats.pc_MetadataEvictions.get());

      // fetch 0d,0md=0 again, evicting 16d,16md=32
      m.cacheRead(P0, 0, 1);
      assertEquals(3, p.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(2, p.stats.pc_MetadataEvictions.get());
   }

   @Test
   public void test_Writes() {
      Processor<RadishLine> p = m.getProc(P0);

      // bring 0d,0md=0 into same set
      m.cacheWrite(P0, 0, 1);
      assertEquals(1, p.stats.pc_l1md.pc_WriteMisses.get());
      assertEquals(0, p.stats.pc_MetadataEvictions.get());

      // evict 0d and 0md=0
      m.cacheWrite(P0, L1_CACHE_SIZE, 1);
      assertEquals(2, p.stats.pc_l1md.pc_WriteMisses.get());
      assertEquals(1, p.stats.pc_MetadataEvictions.get());

      // verify eviction
      m.cacheWrite(P0, 0, 1);
      assertEquals(3, p.stats.pc_l1md.pc_WriteMisses.get());
      assertEquals(2, p.stats.pc_MetadataEvictions.get());
   }

   @Test
   public void test_CrossLineAccess() {
      Processor<RadishLine> p = m.getProc(P0);

      // this crosses a data line boundary, so it definitely crosses a md line boundary
      m.cacheRead(P0, 1, LINE_SIZE);
      assertEquals(2, p.stats.pc_l1d.pc_ReadMisses.get());
      assertEquals(3, p.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(0, p.stats.pc_l1md.pc_ReadHits.get());
      assertEquals(0, p.stats.pc_MetadataEvictions.get());

      // similar thing for writes; no overlapping data or md with previous access
      m.cacheWrite(P0, 1 + (LINE_SIZE * 2), LINE_SIZE);
      assertEquals(2, p.stats.pc_l1d.pc_WriteMisses.get());
      assertEquals(3, p.stats.pc_l1md.pc_WriteMisses.get());
      assertEquals(0, p.stats.pc_l1md.pc_ReadHits.get());
      assertEquals(3, p.stats.pc_MetadataEvictions.get());
   }

   @Test
   public void test_realCacheCrossLine() {
      final int REAL_LINE_SIZE = 64;
      MemorySystemConstants.unsafeSetLineSize(REAL_LINE_SIZE);
      Machine.MachineParams<RadishLine> params = new DefaultTestingMachineParams() {
         @Override
         int numProcessors() {
            return 2;
         }
         
         @Override
         int bitsPerLocalEpochOffset() {
            return 7;
         }
         
         @Override
         CacheConfiguration<RadishLine> l1config() {
            return new CacheConfiguration<RadishLine>() {
               {
                  cacheSize = 65536;
                  lineSize = REAL_LINE_SIZE;
                  assoc = 8;
                  level = CacheLevel.L1;
               }
            };
         }
         
         @Override
         boolean useL3() {
            return true;
         }

         @Override
         CacheConfiguration<RadishLine> l3config() {
            return new CacheConfiguration<RadishLine>() {
               {
                  cacheSize = 64;
                  lineSize = REAL_LINE_SIZE;
                  assoc = 1;
                  level = CacheLevel.L3;
               }
            };
         }
         
         @Override
         boolean useWholeLineSoftwareChecks() {
            return false;
         }
      };
      Machine<RadishLine> machine = new Machine<RadishLine>(params);

      machine.cacheRead(P0, 0x7fffc2afb26fL, 8);
      assertEquals(1, machine.getProc(P0).stats.pc_l1d.pc_ReadMisses.get());
      
      // restore line size
      MemorySystemConstants.unsafeSetLineSize(LINE_SIZE);
   }

   @Test
   public void test_MetadataSize() {
      Processor<RadishLine> p = m.getProc(P0);

      // fetch 0d and 0md=0 into same set
      m.cacheRead(P0, 0, 1);
      assertEquals(1, p.stats.pc_l1d.pc_ReadMisses.get());
      assertEquals(1, p.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(0, p.stats.pc_MetadataEvictions.get());

      // way#: MRU, LRU
      // way0: 0md=0, 0d
      // way1: inv, inv

      // spatial-locality hit on 0d and 0md=0
      m.cacheWrite(P0, (LINE_SIZE - 1) / 2, 1);
      assertEquals(1, p.stats.pc_l1d.pc_WriteHits.get());
      assertEquals(1, p.stats.pc_l1md.pc_WriteHits.get());
      assertEquals(0, p.stats.pc_l1md.pc_WriteMisses.get());
      assertEquals(0, p.stats.pc_MetadataEvictions.get());

      // way0: 0md=0, 0d
      // way1: inv, inv

      // with 2:1 md:d mapping, this is a d hit and md miss
      m.cacheWrite(P0, LINE_SIZE - 1, 1);
      assertEquals(2, p.stats.pc_l1d.pc_WriteHits.get());
      assertEquals(1, p.stats.pc_l1md.pc_WriteMisses.get());
      assertEquals(0, p.stats.pc_MetadataEvictions.get());

      // way0: 0d, 0md=0
      // way1: 3md=6, inv

      // evict some stuff: fetch 4d and 4md=8
      m.cacheRead(P0, LINE_SIZE, 1);
      assertEquals(2, p.stats.pc_l1d.pc_ReadMisses.get());
      assertEquals(2, p.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(1, p.stats.pc_l1md.pc_WriteMisses.get());
      assertEquals(1, p.stats.pc_MetadataEvictions.get());

      // way0: 4md=8, 0d ; evicted: 0md=0
      // way1: 4d, 3md=6

   }

   @Test
   public void test_ClearMetadataFlagOnEviction() {
      Processor<RadishLine> p = m.getProc(P0);

      // we want data and md in different sets
      m.cacheRead(P0, LINE_SIZE / 2, 1);
      assertEquals(1, p.stats.pc_l1d.pc_ReadMisses.get());
      assertEquals(1, p.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(0, p.stats.pc_MetadataEvictions.get());

      // way#: MRU, LRU
      // way0: 2d, inv
      // way1: 2md=4, inv

      // another data-and-md-in-different-sets address pair
      m.cacheRead(P0, (LINE_SIZE / 2) + L1_CACHE_SIZE, 1);
      assertEquals(2, p.stats.pc_l1d.pc_ReadMisses.get());
      assertEquals(2, p.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(0, p.stats.pc_MetadataEvictions.get());

      // way0: 18d, 2d
      // way1: 18md=36, 2md=4

      // recycle one of the lines from way 1 as data, to ensure we clear md flags appropriately
      // we use a read request because md lines should always be in Modified state
      m.cacheRead(P0, LINE_SIZE, 1);
      assertEquals(3, p.stats.pc_l1d.pc_ReadMisses.get());
      assertEquals(3, p.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(1, p.stats.pc_MetadataEvictions.get());

      // way0: 4md=8, 18d
      // way1: 4d, 18md=36

   }

   @Test
   public void test_EpochRollover() {
      Processor<RadishLine> p = m.getProc(P0);

      // check initial state
      assertEquals(EpochBase.INITIAL_EPOCH_BASE, p.localEpochBase);
      assertEquals(EpochOffset.INITIAL_EPOCH_OFFSET, p.epochOffsetManager.current());

      // read 0d and 0md=0, during epoch 1
      m.cacheRead(P0, L1_CACHE_SIZE, 1);
      assertEquals(1, p.stats.pc_l1d.pc_ReadMisses.get());
      assertEquals(1, p.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(0, p.stats.pc_MetadataEvictions.get());
      assertEquals(0, p.stats.pc_EpochRollovers.get());

      // way#: MRU, LRU
      // way0: 0md=0, 0d
      // way1: inv, inv

      m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xDEADBEEF); // causes rollover
      assertEquals(1, p.stats.pc_EpochRollovers.get());
      assertEquals(new EpochBase(1), p.localEpochBase);
      assertEquals(EpochOffset.INITIAL_EPOCH_OFFSET, p.epochOffsetManager.current());

      // 0md=0 must be evicted
      assertEquals(1, p.stats.pc_MetadataEvictions.get());

      m.cacheRead(P0, L1_CACHE_SIZE, 1);
      assertEquals(1, p.stats.pc_l1d.pc_ReadHits.get());
      assertEquals(1, p.stats.pc_l1d.pc_ReadMisses.get());
      assertEquals(0, p.stats.pc_l1md.pc_ReadHits.get());
      assertEquals(2, p.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(1, p.stats.pc_MetadataEvictions.get());
      assertEquals(1, p.stats.pc_EpochRollovers.get());

      m.syncOp(P0, SynchronizationOperationType.SOURCE, true, 0xDEADBEEF); // causes another rollover
      assertEquals(2, p.stats.pc_EpochRollovers.get());

      assertEquals(new EpochBase(2), p.localEpochBase);
      assertEquals(EpochOffset.INITIAL_EPOCH_OFFSET, p.epochOffsetManager.current());

      // 0md=0 must be evicted again
      assertEquals(2, p.stats.pc_MetadataEvictions.get());
   }

   @Test
   public void test_softwareMetadata_RememberSingleThreadedAccessesNoRollover() {
      Processor<RadishLine> p = m.getProc(P0);

      // since there are no previous accesses, hw owns the md now
      m.cacheWrite(P0, 0, 1);
      assertEquals(1, p.stats.pc_SwVcChecksTotal.get());

      // evict 0md=0
      for ( int i = 1; i <= L1_ASSOC; i++ ) {
         m.cacheRead(P0, L1_CACHE_SIZE * i, 1);
         assertEquals(i, p.stats.pc_MetadataEvictions.get());
         assertEquals(1 + i, p.stats.pc_SwVcChecksTotal.get());
      }

      // bring it back, and set local write epoch
      m.cacheRead(P0, 0, 1);
      assertEquals(2 + L1_ASSOC, p.stats.pc_SwVcChecksTotal.get());

      // no sw check because we are the last writer
      m.cacheWrite(P0, 0, 1);
      assertEquals(2 + L1_ASSOC, p.stats.pc_SwVcChecksTotal.get());
   }

   @Test
   public void test_softwareMetadata_RememberSingleThreadedAccessesWithRollover() {
      Processor<RadishLine> p = m.getProc(P0);
      UnownedMetadataByteAddress umba0 = new UnownedMetadataByteAddress(0);

      // since there are no previous accesses, hw owns the md now
      m.cacheWrite(P0, 0, 1);
      assertEquals(1, p.stats.pc_SwVcChecksTotal.get());

      // evict 0md=0
      for ( int i = 1; i <= L1_ASSOC; i++ ) {
         m.cacheRead(P0, L1_CACHE_SIZE * i, 1);
         assertEquals(1 + i, p.stats.pc_SwVcChecksTotal.get());
      }

      // causes rollover, so we can't represent our previous write
      m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0xDEADBEEF);

      // bring it back; since we're totally-ordered wrt all previous accesses, md lives in hw now
      m.cacheRead(P0, 0, 1);
      assertEquals(2 + L1_ASSOC, p.stats.pc_SwVcChecksTotal.get());
      assertEquals( InHardwareStatus.ALL_LEAF_READS, p.inHardwareStatus.get(umba0) );

      m.cacheWrite(P0, 0, 1);
      assertEquals(2 + L1_ASSOC, p.stats.pc_SwVcChecksTotal.get());
   }
   
   @Test
   public void test_WholeLineSwCheck() {
      Processor<RadishLine> p = m.getProc(P0);
      UnownedMetadataByteAddress umba0 = new UnownedMetadataByteAddress(0);

      // since there are no previous accesses, hw owns the md now
      m.cacheWrite(P0, 0, 1);
      assertEquals(1, p.stats.pc_LocalWriteViolatesLocalPerms.get());
      assertEquals(1, p.stats.pc_SwVcChecksTotal.get());
      assertEquals(InHardwareStatus.EVERYTHING, p.inHardwareStatus.get(umba0));

      // whole-line sw check means we should get write perms to the rest of the *md* line
      for ( int i = 1; i < MemorySystemConstants.DATA_BYTES_PER_METADATA_LINE(RadishLine.Compacted.EXPANDED); i++ ) {
         UnownedMetadataByteAddress umba = new UnownedMetadataByteAddress(i);
         m.cacheWrite(P0, i, 1);
         assertEquals(1, p.stats.pc_LocalWriteViolatesLocalPerms.get());
         assertEquals(1, p.stats.pc_SwVcChecksTotal.get());
         assertEquals(InHardwareStatus.EVERYTHING, p.inHardwareStatus.get(umba));
      }
   }

}
